SystemOrganization addCategory: #'AncientCompiler-Debugger'!

MethodMapper subclass: #ACMethodMapper
	instanceVariableNames: 'timestamp methodReference methodNode abstractSourceRanges sortedSourceMap blockExtentsToTempRefs startpcsToTempRefs'
	classVariableNames: 'MapCache MapCacheEntries'
	poolDictionaries: ''
	category: 'AncientCompiler-Debugger'!

!ACMethodMapper class methodsFor: 'debugger support' stamp: 'EstebanAllende 12/13/2011 14:08'!
cacheDebugMap: aDebuggerMethodMap forMethod: aCompiledMethod
	MapCache finalizeValues.
	[MapCache size >= MapCacheEntries] whileTrue:
		[| mapsByAge |
		 mapsByAge := MapCache keys asSortedCollection:
							[:m1 :m2|
							(MapCache at: m1) timestamp
							< (MapCache at: m2) timestamp].
		mapsByAge notEmpty ifTrue: "There be race conditions and reentrancy issues here"
			[MapCache removeKey: mapsByAge last]].

	^MapCache
		at: aCompiledMethod
		put: aDebuggerMethodMap! !

!ACMethodMapper class methodsFor: 'cleanup' stamp: 'EstebanAllende 12/13/2011 14:08'!
cleanUp
	"DebuggerMethodMap cleanUp"

	self voidMapCache! !

!ACMethodMapper class methodsFor: 'instance creation' stamp: 'EstebanAllende 12/13/2011 14:08'!
forMethod: aMethod "<CompiledMethod>"
	"Answer a DebuggerMethodMap suitable for debugging activations of aMethod.
	 Answer an existing instance from the cache if it exists, cacheing a new one if required."
	^MapCache
		at: aMethod
		ifAbsent: [self
					cacheDebugMap:
						(self
							forMethod: aMethod
							methodNode: aMethod methodNode)
					forMethod: aMethod]! !

!ACMethodMapper class methodsFor: 'instance creation' stamp: 'EstebanAllende 12/13/2011 14:08'!
forMethod: aMethod "<CompiledMethod>" methodNode: methodNode "<MethodNode>"
	"Uncached instance creation method for private use or for tests.
	 Please consider using forMethod: instead."
	^self new
		forMethod: aMethod
		methodNode: methodNode! !

!ACMethodMapper class methodsFor: 'class initialization' stamp: 'EstebanAllende 12/13/2011 14:08'!
initialize
	self cleanUp! !

!ACMethodMapper class methodsFor: 'class initialization' stamp: 'EstebanAllende 12/13/2011 14:08'!
voidMapCache
	MapCache := WeakIdentityKeyDictionary new.
	MapCacheEntries := 16! !

!ACMethodMapper methodsFor: 'source mapping' stamp: 'EstebanAllende 12/13/2011 14:08'!
abstractPCFor: concretePC when: contextIsActive
	<PRIVATE>
	"If the context is the actve context (is at the hot end of the stack)
	 then its pc is the current pc.  But if the context isn't, because it is
	suspended sending a message, then its current pc is the previous pc"
	^self method abstractPCForConcretePC:
		(contextIsActive 
			ifTrue: [concretePC]
			ifFalse: [(self method pcPreviousTo: concretePC) ifNil: [concretePC]])! !

!ACMethodMapper methodsFor: 'source mapping' stamp: 'EstebanAllende 12/13/2011 14:08'!
abstractSourceMap
	<PRIVATE>
	"Answer with a Dictionary of abstractPC <Integer> to sourceRange <Interval>."
	| theMethodToScan rawSourceRanges concreteSourceRanges abstractPC scanner client |
	abstractSourceRanges ifNotNil:
		[^abstractSourceRanges].
	"If the methodNode hasn't had a method generated it doesn't have pcs set in its
	 nodes so we must generate a new method and might as well use it for scanning."
	methodNode rawSourceRangesAndMethodDo:
		[:ranges :method|
		 rawSourceRanges := ranges.
		 theMethodToScan := method].
	concreteSourceRanges := Dictionary new.
	rawSourceRanges keysAndValuesDo:
		[:node :range|
		node pc ~= 0 ifTrue:
			[concreteSourceRanges at: node pc put: range]].
	abstractPC := 1.
	abstractSourceRanges := Dictionary new.
	scanner := InstructionStream on: theMethodToScan.
	client := InstructionClient new.
	[(concreteSourceRanges includesKey: scanner pc) ifTrue:
		[abstractSourceRanges at: abstractPC put: (concreteSourceRanges at: scanner pc)].
	 abstractPC := abstractPC + 1.
	 scanner interpretNextInstructionFor: client.
	 scanner atEnd] whileFalse.
	^abstractSourceRanges! !

!ACMethodMapper methodsFor: 'initialize-release' stamp: 'EstebanAllende 12/13/2011 14:08'!
forMethod: aMethod "<CompiledMethod>" methodNode: theMethodNode "<MethodNode>"
	<INIT>
	methodReference := WeakArray with: aMethod.
	methodNode := theMethodNode.
	self markRecentlyUsed! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
markRecentlyUsed
	<PRIVATE>
	timestamp := Time totalSeconds! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
method
	<UNKNOWN>
	^methodReference at: 1! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
namedTempAt: index in: aContext
	"Answer the value of the temp at index in aContext where index is relative
	 to the array of temp names answered by tempNamesForContext:"
	^self
		privateTempAt: index
		in: aContext
		startpcsToBlockExtents: aContext method startpcsToBlockExtents! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
namedTempAt: index put: aValue in: aContext
	"Assign the value of the temp at index in aContext where index is relative
	 to the array of temp names answered by tempNamesForContext:.
	 If the value is a copied value we also need to set it along the lexical chain."
	^self
		privateTempAt: index
		in: aContext
		put: aValue
		startpcsToBlockExtents: aContext method startpcsToBlockExtents! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateDereference: tempReference in: aContext
	<PRIVATE>
	"Fetch the temporary with reference tempReference in aContext.
	 tempReference can be
		integer - direct temp reference
		#( indirectionVectorIndex tempIndex ) - remote temp in indirectionVector at index
		#( outer. temp reference ) - a temp reference in an outer context."
	^tempReference isInteger
		ifTrue: [aContext tempAt: tempReference]
		ifFalse:
			[tempReference first == #outer
				ifTrue: [self privateDereference: tempReference last
							in: aContext outerContext]
				ifFalse: [(aContext tempAt: tempReference first)
							at: tempReference second]]! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateDereference: tempReference in: aContext put: aValue
	<PRIVATE>
	"Assign the temporary with reference tempReference in aContext.
	 tempReference can be
		integer - direct temp reference
		#( indirectionVectorIndex tempIndex ) - remote temp in indirectionVector at index
		#( outer. temp reference ) - a temp reference in an outer context."
	^tempReference isInteger
		ifTrue: [aContext tempAt: tempReference put: aValue]
		ifFalse:
			[tempReference first == #outer
				ifTrue: [self privateDereference: tempReference last
							in: aContext outerContext
							put: aValue]
				ifFalse: [(aContext tempAt: tempReference first)
							at: tempReference second
							put: aValue]]! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateIsOuter: anObject 
	<PRIVATE>
	^anObject last isArray and: [anObject last first == #outer]! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateTempAt: index in: aContext put: aValue startpcsToBlockExtents: theContextsStartpcsToBlockExtents
	| nameRefPair |
	<PRIVATE>
	nameRefPair := (self privateTempRefsForContext: aContext
						 startpcsToBlockExtents: theContextsStartpcsToBlockExtents)
						at: index
						ifAbsent: [aContext errorSubscriptBounds: index].
	^self privateDereference: nameRefPair last in: aContext put: aValue! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateTempAt: index in: aContext startpcsToBlockExtents: theContextsStartpcsToBlockExtents
	| nameRefPair |
	<PRIVATE>
	nameRefPair := (self privateTempRefsForContext: aContext
						 startpcsToBlockExtents: theContextsStartpcsToBlockExtents)
						at: index
						ifAbsent: [aContext errorSubscriptBounds: index].
	^self privateDereference: nameRefPair last in: aContext! !

!ACMethodMapper methodsFor: 'private' stamp: 'EstebanAllende 12/13/2011 14:08'!
privateTempRefsForContext: aContext startpcsToBlockExtents: theContextsStartpcsToBlockExtents
	<PRIVATE>
	"Answer the sequence of temps in scope in aContext in the natural order,
	 outermost arguments and temporaries first, innermost last.  Each temp is
	 a pair of the temp's name followed by a reference.  The reference can be
		integer - index of temp in aContext
		#( indirectionVectorIndex tempIndex ) - remote temp in indirectionVector at index in aContext
		#( outer. temp reference ) - a temp reference in an outer context."
	blockExtentsToTempRefs ifNil:
		[blockExtentsToTempRefs := (aContext method holdsTempNames
										ifTrue: [aContext method]
										ifFalse: [methodNode]) blockExtentsToTempsMap.
		 startpcsToTempRefs := Dictionary new].
	^startpcsToTempRefs
		at: aContext startpc
		ifAbsentPut:
			[| localRefs |
			 localRefs := blockExtentsToTempRefs at: (theContextsStartpcsToBlockExtents at: aContext startpc).
			 aContext outerContext
				ifNil: [localRefs]
				ifNotNil:
					[:outer| | outerTemps |
					"Present temps in the order outermost to innermost left-to-right, but replace
					 copied outermost temps with their innermost copies"
					 outerTemps := (self
										privateTempRefsForContext: outer
										startpcsToBlockExtents: theContextsStartpcsToBlockExtents) collect:
						[:outerPair|
						localRefs
							detect: [:localPair| outerPair first = localPair first]
							ifNone: [{ outerPair first. { #outer. outerPair last } }]].
					outerTemps,
					 (localRefs reject: [:localPair| outerTemps anySatisfy: [:outerPair| localPair first = outerPair first]])]]! !

!ACMethodMapper methodsFor: 'source mapping' stamp: 'EstebanAllende 12/13/2011 14:08'!
rangeForPC: concretePC contextIsActiveContext: contextIsActive
  "Answer the indices in the source code for the supplied pc."
	| pc |
	pc := self abstractPCFor: concretePC when: contextIsActive.
	self sortedSourceMap first key > pc ifTrue: [^self sortedSourceMap first value].
	^self abstractSourceMap at: pc ifAbsent:
    		[(self sortedSourceMap 
			findBinary: [:assoc | pc - assoc key ] 
			ifNone: [ | end |
					end := self sortedSourceMap last value last.
					nil -> (end +1 to: end)]) value]! !

!ACMethodMapper methodsFor: 'source mapping' stamp: 'EstebanAllende 12/13/2011 14:08'!
sortedSourceMap
	<PRIVATE>
	"Answer a sorted collection of associations, pcRangeStart -> pcRangeInterval "
	^ sortedSourceMap
		ifNil: [sortedSourceMap := self abstractSourceMap associations sorted]! !

!ACMethodMapper methodsFor: 'source mapping' stamp: 'EstebanAllende 12/13/2011 14:08'!
sourceText
	<UNKNOWN>
	self method ifNotNil:
		[:method|
		method holdsTempNames ifTrue:
			[^method
				getSourceFor: (method selector ifNil: [method defaultSelector])
				in: method methodClass]].
	^methodNode sourceText! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
tempNamesForContext: aContext
	"Answer an Array of all the temp names in scope in aContext starting with
	 the home's first local (the first argument or first temporary if no arguments)."
	^(self
		privateTempRefsForContext: aContext
		startpcsToBlockExtents: aContext method startpcsToBlockExtents) collect:
			[:pair| pair first]! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
tempNamesScopedForContext: aContext
	"Answer an Array of all the temp names in scope in aContext starting with
	 the home's first local (the first argument or first temporary if no arguments)."
	^((self
		privateTempRefsForContext: aContext
		startpcsToBlockExtents: aContext method startpcsToBlockExtents) reject: [:pair | self privateIsOuter: pair] )
		collect:
			[:pair| pair first]! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
tempsAndValuesForContext: aContext
	"Return a string of the temporary variabls and their current values"
	| aStream |
	aStream := (String new: 100) writeStream.
	(self tempNamesForContext: aContext) doWithIndex:
		[:title :index |
		 aStream nextPutAll: title; nextPut: $:; space; tab.
		 aContext print: (self namedTempAt: index in: aContext) on: aStream.
		 aStream cr].
	^aStream contents! !

!ACMethodMapper methodsFor: 'accessing' stamp: 'EstebanAllende 12/13/2011 14:08'!
timestamp
	<PRIVATE>
	^timestamp! !
ACMethodMapper initialize!
